# include <limits>
# include <iostream>

# include <dbProperties.H>

# include <compute_exchange.H>

double compute_exchange_by_product(const unsigned long long & product_id,
                                   const char * const year,
                                   const double & exchange_type,
                                   const char * const host,
                                   const char * const port,
                                   const char * const dbname,
                                   const char * const username,
                                   const char * const password)
{
  DBProperties & dbProp = DBProperties::getInstance();

  dbProp.setHost(host);
  dbProp.setPort(port);
  dbProp.setDbname(dbname);
  dbProp.setUser(username);
  dbProp.setPassword(password);

  IO_Manager io_manager;

  const int levels_up = std::numeric_limits<int>::max();

  const int levels_down = 0;

  char * const xml_root = "temp.xml";

  io_manager.create_xml_root_file(year, product_id, levels_up, levels_down,
                                  xml_root);

  double ret_val = 0.0;

  try
  {
    Graph graph;

    io_manager.build_graph(xml_root, graph);

    for (Graph::Node_Iterator it(graph); it.has_current(); it.next())
      {
        Graph::Node * node = it.get_current();

        if (node->get_info()->get_type() != IMPORTED_PRODUCT)
          continue;

        for (Graph::Node_Arc_Iterator arc_it(node); arc_it.has_current();
             arc_it.next())
          {
            Graph::Arc * arc = arc_it.get_current();

            IP_Relationship & ip_rel = arc->get_info();

            ret_val += (ip_rel.get_bought_quantity() *
                        ip_rel.get_purchase_price()) / exchange_type;
          }
      }
  }
  catch (const std::exception & e)
  {
    std::cout << "Excepción capturada: " << e.what() << std::endl;
  }

  remove(xml_root);

  return ret_val;  
}

