/*
  Copyright (C) 2014
  Alejandro Mujica (amujica@cenditel.gob.ve)
  José Ruiz (jruiz@cenditel.gob.ve)
  Julie Vera (jvera@cenditel.gob.ve)
 
  CENDITEL Fundación Centro Nacional de Desarrollo e Investigación en
  Tecnologías Libres
 
  Este programa es software libre; Usted puede usarlo bajo los términos de la
  licencia de software GPL versión 2.0 de la Free Software Foundation.
 
  Este programa se distribuye con la esperanza de que sea útil, pero SIN
  NINGUNA GARANTÍA; tampoco las implícitas garantías de MERCANTILIDAD o
  ADECUACIÓN A UN PROPÓSITO PARTICULAR.
  Consulte la licencia GPL para más detalles. Usted debe recibir una copia
  de la GPL junto con este programa; si no, escriba a la Free Software
  Foundation Inc. 51 Franklin Street,5 Piso, Boston, MA 02110-1301, USA.
*/

/*
  Autor:             Alejandro J. Mujica
  Fecha de creación: 12/06/2014
  Este archivo contiene la implementación de las diversas consultas
  a base de datos efectuadas.
*/

# include <queries.H>

# include <sstream>
# include <stdexcept>

# include <dbQuery.H>
# include <autoConnection.H>
# include <strQuery.H>

const std::string DB::PRODUCT_TABLE_NAME = "produccion_producto";
const std::string DB::INPUT_TABLE_NAME = "produccion_insumo";
const std::string DB::PLANT_TABLE_NAME = "unidadecon_subunidad_economica";
const std::string DB::COMPANY_TABLE_NAME = "unidadecon_unidad_economica";
const std::string DB::COMPANY_ACTIVITY_CIIU_TABLE_NAME =
  "unidadecon_actividad_ciiu";
const std::string DB::PRODUCT_INPUT_TABLE_NAME = "produccion_producto_t_insumo";
const std::string DB::MEASUREMENT_UNIT_TABLE_NAME = "comun_unida_medida";
const std::string DB::INPUT_SUPPLIER_TABLE_NAME =
  "cmproveedores_proveedorinsumo";
const std::string DB::SUPPLIER_TABLE_NAME = "cmproveedores_proveedor";
const std::string DB::PRODUCTION_QUANTITIES_TABLE_NAME =
    "produccion_produccion_precios_r";
const std::string DB::ACQ_PRICE_TABLE_NAME = "produccion_consumo_precios_r";
const std::string DB::COUNTRY_TABLE_NAME = "comun_pais";
 const std::string DB::PARISH_TABLE_NAME = "comun_parroquia";
const std::string DB::TOWN_TABLE_NAME = "comun_municipio";
const std::string DB::STATE_TABLE_NAME = "comun_entidad";
const std::string DB::COMPANY_OCCUPANCY_TABLE_NAME =
    "unidadecon_personal_ocupado";

const std::string DB::ID = "id";
const std::string DB::NAME = "nombre";
const std::string DB::COMPANY_NAME = "razon_social";
const std::string DB::TECHNICAL_SPECIFICATION = "esp_tec";
const std::string DB::TARIFF_CODE = "codigo_aran_id";
const std::string DB::MEASUREMENT_UNIT_ID = "unidad_medida_id";
const std::string DB::TRADEMARK = "marca";
const std::string DB::PLANT_ID = "planta_id_id";
const std::string DB::CODE = "codigo";
const std::string DB::PARISH_CODE = "parroquia_codigo";
const std::string DB::TOWN_CODE = "municipio_codigo";
const std::string DB::STATE_CODE = "entidad_codigo";
const std::string DB::COMPANY_RIF = "unidad_economica_rif";
const std::string DB::UE_RIF = "ue_rif";
const std::string DB::RIF = "rif";
const std::string DB::NATIONALITY = "nacional";
const std::string DB::WORKDAY = "horas_turnos";
const std::string DB::ACQ_PRICE = "cost_adq";
const std::string DB::NAC_ACQ_PRICE = "prec_adq_nac";
const std::string DB::CIIU_ID = "ciiu_id";
const std::string DB::ACTIVE = "activo";
const std::string DB::MAIN = "principal";
const std::string DB::INPUT_ID = "insumo_id_id";
const std::string DB::PRODUCT_ID = "producto_id_id";
const std::string DB::SUPPLIER_ID = "proveedor_id_id";
const std::string DB::PRODUCED_QUANTITY = "cant_producido";
const std::string DB::PRICE = "precio_venta";
const std::string DB::BOUGHT_QUANTITY = "cantidad_comprada_r";
const std::string DB::QUANTITY = "cantidad";
const std::string DB::ORIG_COUNTRY_ID= "pais_origen_id";
const std::string DB::PROC_COUNTRY_ID = "pais_procedencia_id";
const std::string DB::YEAR_R = "anho_r";
const std::string DB::YEAR = "anho";
const std::string DB::USED_CAPACITY = "capacidad_utilizada";
const std::string DB::INSERTION_DATE = "fecha_insercion";
const std::string DB::OPERATING = "operativo";
const std::string DB::ADMINISTRATIVE = "administrativo";

std::string single_quote(const std::string & str)
{
  return "'" + str + "'";
}

std::string double_quote(const std::string & str)
{
  return "\"" + str + "\"";
}

std::string concat(const std::string & table, const std::string & field,
                   const char & connector)
{
  return table + connector + field;
}

std::string as(const std::string & field, const std::string & alias)
{
  return field + " AS " + alias;
}

std::string concatas(const std::string & table, const std::string & field,
                     const std::string & alias)
{
  return as(concat(table, field), alias);
}

std::string concatas(const std::string & table, const std::string & field)
{
  return concatas(table, field, concat(table, field, '_'));
}

std::string op(const std::string & attr, const std::string & value,
               const std::string & oper)
{
  return attr + " " + oper + " " + value;
}

std::string id_to_str(const db_id_t & id)
{
  std::stringstream sstr;
  sstr << id;
  return sstr.str();
}

std::string between(const std::string & field, const std::string & beg,
                    const std::string & end)
{
  return field + " BETWEEN " + single_quote(beg) + " AND " + single_quote(end);
}

std::string in_year(const std::string & field, const std::string & year)
{
  std::string beg = year + "-01-01";
  std::string end = year + "-12-31";
  return between(field, beg, end);
}

Product load_product_by_id(const db_id_t & id)
{
  AutoConnection conn;
  DBQuery query(conn);
  StrQuery str_query;

  str_query.addSelect(concatas(DB::PRODUCT_TABLE_NAME, DB::ID));
  str_query.addSelect(concatas(DB::PRODUCT_TABLE_NAME, DB::NAME));
  str_query.addSelect(concatas(DB::PRODUCT_TABLE_NAME, DB::TARIFF_CODE));
  str_query.addSelect(concatas(DB::MEASUREMENT_UNIT_TABLE_NAME, DB::NAME));
  str_query.addSelect(concatas(DB::PLANT_TABLE_NAME, DB::ID));
  str_query.addSelect(concatas(DB::COMPANY_TABLE_NAME, DB::RIF));
  str_query.addSelect(concatas(DB::COMPANY_TABLE_NAME, DB::COMPANY_NAME));
  str_query.addSelect(concatas(DB::STATE_TABLE_NAME, DB::NAME));

  str_query.addFrom(DB::PRODUCT_TABLE_NAME);
  str_query.addFrom(DB::MEASUREMENT_UNIT_TABLE_NAME);
  str_query.addFrom(DB::PLANT_TABLE_NAME);
  str_query.addFrom(DB::COMPANY_TABLE_NAME);
  str_query.addFrom(DB::STATE_TABLE_NAME);
  str_query.addFrom(DB::TOWN_TABLE_NAME);
  str_query.addFrom(DB::PARISH_TABLE_NAME);

  str_query.addWhere(op(concat(DB::PRODUCT_TABLE_NAME, DB::ID),
                        id_to_str(id)));
  str_query.addWhere(op(concat(DB::MEASUREMENT_UNIT_TABLE_NAME, DB::ID),
                        concat(DB::PRODUCT_TABLE_NAME,
                               DB::MEASUREMENT_UNIT_ID)));
  str_query.addWhere(op(concat(DB::PLANT_TABLE_NAME, DB::ID),
                        concat(DB::PRODUCT_TABLE_NAME, DB::PLANT_ID)));
  str_query.addWhere(op(concat(DB::PLANT_TABLE_NAME, DB::ID),
                        concat(DB::PRODUCT_TABLE_NAME, DB::PLANT_ID)));
  str_query.addWhere(op(concat(DB::COMPANY_TABLE_NAME, DB::RIF),
                        concat(DB::PLANT_TABLE_NAME, DB::UE_RIF)));
  str_query.addWhere(op(concat(DB::COMPANY_TABLE_NAME, DB::PARISH_CODE),
                        concat(DB::PARISH_TABLE_NAME, DB::CODE)));
  str_query.addWhere(op(concat(DB::PARISH_TABLE_NAME, DB::TOWN_CODE),
                        concat(DB::TOWN_TABLE_NAME, DB::CODE)));
  str_query.addWhere(op(concat(DB::TOWN_TABLE_NAME, DB::STATE_CODE),
                        concat(DB::STATE_TABLE_NAME, DB::CODE)));

  if (not query.exec(str_query))
    {
      std::string msg = "Cannot execute query: " + std::string(str_query);
      throw std::logic_error(msg);
    }

  Product product;

  if (not query.next())
    return product;

  product.set_company_name(
    query.getValue(concat(DB::COMPANY_TABLE_NAME, DB::COMPANY_NAME, '_')));
  product.set_company_rif(
    query.getValue(concat(DB::COMPANY_TABLE_NAME, DB::RIF, '_')));
  product.set_company_location(
    query.getValue(concat(DB::STATE_TABLE_NAME, DB::NAME, '_')));
  product.set_id(
    std::atol(query.getValue(concat(DB::PRODUCT_TABLE_NAME, DB::ID, '_'))));
  product.set_name(query.getValue(concat(DB::PRODUCT_TABLE_NAME,
                                         DB::NAME, '_')));
  product.set_tariff_code(
    query.getValue(concat(DB::PRODUCT_TABLE_NAME, DB::TARIFF_CODE, '_')));
  product.set_measurement_unit(
    query.getValue(concat(DB::MEASUREMENT_UNIT_TABLE_NAME, DB::NAME, '_')));

  return product;
}

void list_input_ids(const db_id_t & product_id, const std::string & year,
                    List<db_id_t> & input_ids)
{
  AutoConnection conn;
  DBQuery query(conn);
  StrQuery str_query;

  str_query.addSelect(DB::INPUT_ID);
  str_query.addFrom(DB::PRODUCT_INPUT_TABLE_NAME);
  str_query.addWhere(op(DB::PRODUCT_ID, id_to_str(product_id)));
  str_query.addWhere(op(DB::QUANTITY, "0", "<>"));
  str_query.addWhere(op(DB::YEAR, year));

  if (not query.exec(str_query))
    {
      std::string msg = "Cannot execute query: " + std::string(str_query);
      throw std::logic_error(msg);
    }

  while (query.next())
    input_ids.append(std::atol(query.getValue(DB::INPUT_ID)));
}

void list_inputs(const db_id_t & input_id, const std::string & year,
                 List<Input> & inputs)
{
  AutoConnection conn;
  DBQuery query(conn);
  StrQuery str_query;

  str_query.addSelect(concatas(DB::INPUT_TABLE_NAME, DB::ID));
  str_query.addSelect(concatas(DB::INPUT_TABLE_NAME, DB::NAME));
  str_query.addSelect(concatas(DB::INPUT_TABLE_NAME, DB::TARIFF_CODE));
  str_query.addSelect(concatas(DB::MEASUREMENT_UNIT_TABLE_NAME, DB::NAME));

  str_query.addFrom(DB::INPUT_TABLE_NAME);
  str_query.addFrom(DB::MEASUREMENT_UNIT_TABLE_NAME);

  str_query.addWhere(op(concat(DB::INPUT_TABLE_NAME, DB::ID),
                               id_to_str(input_id)));
  str_query.addWhere(op(concat(DB::INPUT_TABLE_NAME,
                               DB::MEASUREMENT_UNIT_ID),
                        concat(DB::MEASUREMENT_UNIT_TABLE_NAME, DB::ID)));

  if (not query.exec(str_query))
    {
      std::string msg = "Cannot execute query: " + std::string(str_query);
      throw std::logic_error(msg);
    }

  if (not query.next())
    {
      std::stringstream msg;
      msg << "Input with id = " << input_id << " doesn't exist in database";
      throw std::domain_error(msg.str());
    }

  Input input;

  input.set_id(
    std::atol(query.getValue(concat(DB::INPUT_TABLE_NAME, DB::ID, '_'))));
  input.set_name(query.getValue(concat(DB::INPUT_TABLE_NAME, DB::NAME, '_')));
  input.set_tariff_code(
    query.getValue(concat(DB::INPUT_TABLE_NAME, DB::TARIFF_CODE, '_')));
  input.set_measurement_unit(
    query.getValue(concat(DB::MEASUREMENT_UNIT_TABLE_NAME, DB::NAME, '_')));

  str_query.clear();
  query.clear();

  str_query.addSelect(concatas(DB::SUPPLIER_TABLE_NAME, DB::RIF));
  str_query.addSelect(concatas(DB::SUPPLIER_TABLE_NAME, DB::NAME));
  str_query.addSelect(concatas(DB::SUPPLIER_TABLE_NAME, DB::NATIONALITY));
  str_query.addSelect(concatas(DB::COUNTRY_TABLE_NAME, DB::NAME));

  str_query.addFrom(DB::INPUT_SUPPLIER_TABLE_NAME);
  str_query.addFrom(DB::SUPPLIER_TABLE_NAME);
  str_query.addFrom(DB::COUNTRY_TABLE_NAME);

  str_query.addWhere(op(concat(DB::INPUT_SUPPLIER_TABLE_NAME, DB::YEAR_R),
                        year));
  str_query.addWhere(op(concat(DB::INPUT_SUPPLIER_TABLE_NAME, DB::INPUT_ID),
                     id_to_str(input_id)));
  str_query.addWhere(op(concat(DB::INPUT_SUPPLIER_TABLE_NAME, DB::SUPPLIER_ID),
                        concat(DB::SUPPLIER_TABLE_NAME, DB::ID)));
  str_query.addWhere(op(concat(DB::COUNTRY_TABLE_NAME, DB::ID),
                        concat(DB::SUPPLIER_TABLE_NAME,
                               DB::ORIG_COUNTRY_ID)));

  str_query.addOrderBy(concat(DB::SUPPLIER_TABLE_NAME, DB::RIF));

  if (not query.exec(str_query))
    {
      std::string msg = "Cannot execute query: " + std::string(str_query);
      throw std::logic_error(msg);
    }

  std::string last_rif = "";

  while (query.next())
    {
      std::string rif =
        query.getValue(concat(DB::SUPPLIER_TABLE_NAME, DB::RIF, '_'));

      std::string nationality = query.getValue(concat(DB::SUPPLIER_TABLE_NAME,
                                                      DB::NATIONALITY, '_'));

      if (nationality != "E" and rif == last_rif)
        continue;

      last_rif = rif;

      input.set_company_rif(rif);
      input.set_company_name(
        query.getValue(concat(DB::SUPPLIER_TABLE_NAME, DB::NAME, '_'))
      );
      input.set_company_location(
        query.getValue(concat(DB::COUNTRY_TABLE_NAME, DB::NAME, '_'))
      );
      input.set_nationality(nationality);
      inputs.append(input);
    }
}

void product_from_input(const Input & input, List<Product> & products)
{
  AutoConnection conn;
  DBQuery query(conn);
  StrQuery str_query;

  str_query.addSelect(concatas(DB::PRODUCT_TABLE_NAME, DB::ID));

  str_query.addFrom(DB::INPUT_TABLE_NAME);
  str_query.addFrom(DB::PRODUCT_TABLE_NAME);
  str_query.addFrom(DB::PLANT_TABLE_NAME);

  str_query.addWhere(op(concat(DB::INPUT_TABLE_NAME, DB::ID),
                        id_to_str(input.get_id())));
  str_query.addWhere(op(concat(DB::INPUT_TABLE_NAME, DB::TARIFF_CODE),
                        concat(DB::PRODUCT_TABLE_NAME, DB::TARIFF_CODE)));
  str_query.addWhere(op(concat(DB::PLANT_TABLE_NAME, DB::ID),
                        concat(DB::PRODUCT_TABLE_NAME, DB::PLANT_ID)));
  str_query.addWhere(op(concat(DB::PLANT_TABLE_NAME, DB::UE_RIF),
                        single_quote(input.get_company_rif())));

  if (not query.exec(str_query))
    {
      std::string msg = "Cannot execute query: " + std::string(str_query);
      throw std::logic_error(msg);
    }

  while (query.next())
    products.append(load_product_by_id(
      std::atol(query.getValue(concat(DB::PRODUCT_TABLE_NAME, DB::ID, '_')))));
}

real get_unitarian_request(const db_id_t & input_id,
                           const db_id_t & product_id)
{
  AutoConnection conn;
  DBQuery query(conn);
  StrQuery str_query;

  str_query.addSelect(DB::QUANTITY);

  str_query.addFrom(DB::PRODUCT_INPUT_TABLE_NAME);

  str_query.addWhere(op(DB::PRODUCT_ID, id_to_str(product_id)));
  str_query.addWhere(op(DB::INPUT_ID, id_to_str(input_id)));

  if (not query.exec(str_query))
    {
      std::string msg = "Cannot execute query: " + std::string(str_query);
      throw std::logic_error(msg);
    }

  if (not query.next())
    return 0.0;

  return std::atof(query.getValue(0));
}

real get_acquisition_price(const std::string & nationality,
                           const db_id_t & input_id,
                           const std::string & year)
{
  AutoConnection conn;
  DBQuery query(conn);
  StrQuery str_query;

  std::string price_field = DB::NAC_ACQ_PRICE;

  if (nationality == "E")      
    price_field = DB::ACQ_PRICE;

  str_query.addSelect(price_field);

  str_query.addFrom(DB::ACQ_PRICE_TABLE_NAME);

  str_query.addWhere(op(DB::INPUT_ID, id_to_str(input_id)));
  str_query.addWhere(in_year(DB::YEAR, year));

  if (not query.exec(str_query))
    {
      std::string msg = "Cannot execute query: " + std::string(str_query);
      throw std::logic_error(msg);
    } 

  if (not query.next()) 
    return 0.0;

  return std::atof(query.getValue(0));
}

real get_declared_bought_quantity(const std::string & rif,
                                  const std::string & company_name,
                                  const db_id_t & input_id,
                                  const std::string & year)
{
  AutoConnection conn;

  DBQuery query(conn);

  std::string sum = "SUM(" + concat(DB::INPUT_SUPPLIER_TABLE_NAME,
                                    DB::BOUGHT_QUANTITY) + ")";

  std::string str_query = "SELECT " + sum
                        + " FROM " + DB::INPUT_SUPPLIER_TABLE_NAME + ", "
                                   + DB::SUPPLIER_TABLE_NAME +
                        + " WHERE " + 
                        op(concat(DB::INPUT_SUPPLIER_TABLE_NAME, DB::YEAR_R),
                           year) + " AND " +
                        op(concat(DB::SUPPLIER_TABLE_NAME, DB::ID),
                           concat(DB::INPUT_SUPPLIER_TABLE_NAME,
                                  DB::SUPPLIER_ID)) + " AND " +
                        op(concat(DB::INPUT_SUPPLIER_TABLE_NAME, DB::INPUT_ID),
                                  id_to_str(input_id)) + " AND ";

  std::string group = " GROUP BY ";

  if (rif.size() > 0)
    {
      str_query.append(op(concat(DB::SUPPLIER_TABLE_NAME, DB::RIF),
                          single_quote(rif)));
      group.append(concat(DB::SUPPLIER_TABLE_NAME, DB::RIF));
    }
  else
    {
      str_query.append(op(concat(DB::SUPPLIER_TABLE_NAME, DB::NAME),
                          single_quote(company_name)));
      group.append(concat(DB::SUPPLIER_TABLE_NAME, DB::NAME));
    }

  str_query.append(group);

  if (not query.exec(str_query))
    {
      std::string msg = "Cannot execute query: " + str_query;
      throw std::logic_error(msg);
    }

  if (not query.next())
    return 0.0;

  return std::atof(query.getValue(0));
}

Aleph::pair<real, real>
get_produced_quantity_and_price(const db_id_t & product_id,
                                const std::string & year)
{
  AutoConnection conn;
  DBQuery query(conn);
  StrQuery str_query;

  str_query.addSelect(DB::PRODUCED_QUANTITY);
  str_query.addSelect(DB::PRICE);

  str_query.addFrom(DB::PRODUCTION_QUANTITIES_TABLE_NAME);

  str_query.addWhere(op(DB::PRODUCT_ID, id_to_str(product_id)));
  str_query.addWhere(in_year(DB::YEAR, year));

  if (not query.exec(str_query))
    {
      std::string msg = "Cannot execute query: " + std::string(str_query);
      throw std::logic_error(msg);
    }

  if (not query.next())
    return Aleph::pair<real, real>(0.0, 0.0);

  return Aleph::pair<real, real>(std::atof(query.getValue(0)),
                                 std::atof(query.getValue(1)));
}

void list_product_ids(Good * good, List<Aleph::pair<long, long>> & list)
{
  AutoConnection conn;
  DBQuery query(conn);
  StrQuery str_query;

  const std::string & rif = good->get_company_rif();

  const std::string & tariff_code = good->get_tariff_code();

  str_query.addSelect(concatas(DB::PRODUCT_INPUT_TABLE_NAME, DB::PRODUCT_ID));
  str_query.addSelect(concatas(DB::PRODUCT_INPUT_TABLE_NAME, DB::INPUT_ID));

  str_query.addFrom(DB::SUPPLIER_TABLE_NAME);
  str_query.addFrom(DB::INPUT_SUPPLIER_TABLE_NAME);
  str_query.addFrom(DB::PRODUCT_INPUT_TABLE_NAME);
  str_query.addFrom(DB::INPUT_TABLE_NAME);
  str_query.addFrom(DB::PRODUCT_TABLE_NAME);

  str_query.addWhere(op(concat(DB::SUPPLIER_TABLE_NAME, DB::RIF),
                        single_quote(rif)));
  str_query.addWhere(op(concat(DB::SUPPLIER_TABLE_NAME, DB::ID),
                        concat(DB::INPUT_SUPPLIER_TABLE_NAME,
                               DB::SUPPLIER_ID)));
  str_query.addWhere(op(concat(DB::INPUT_SUPPLIER_TABLE_NAME, DB::INPUT_ID),
                       concat(DB::INPUT_TABLE_NAME, DB::ID)));
  str_query.addWhere(op(concat(DB::INPUT_TABLE_NAME, DB::TARIFF_CODE),
                       single_quote(tariff_code)));
  str_query.addWhere(op(concat(DB::INPUT_SUPPLIER_TABLE_NAME, DB::INPUT_ID),
                         concat(DB::PRODUCT_INPUT_TABLE_NAME, DB::INPUT_ID)));
  str_query.addWhere(op(concat(DB::PRODUCT_INPUT_TABLE_NAME, DB::PRODUCT_ID),
                        concat(DB::PRODUCT_TABLE_NAME, DB::ID)));
  str_query.addWhere(op(concat(DB::PRODUCT_INPUT_TABLE_NAME, DB::QUANTITY),
                               "0", "<>"));

  if (not query.exec(str_query))
    {
      std::string msg = "Cannot execute query: " + std::string(str_query);
      throw std::logic_error(msg);
    }

  while (query.next())
    list.append(Aleph::pair<long, long>(std::atol(query.getValue(0)),
                                        std::atol(query.getValue(1))));
}

void list_requesters(Good * good, List<Aleph::pair<Product, long>> & list)
{
  List<Aleph::pair<long, long>> ids;

  list_product_ids(good, ids);

  for(List <Aleph::pair<long, long>>::Iterator it(ids); it.has_current();
      it.next())
    {
      Aleph::pair<long, long> & p = it.get_current();

      Product product = load_product_by_id(p.first);

      list.append(Aleph::pair<Product, long>(product, p.second));
    }
}

real get_product_used_capacity(const db_id_t & product_id,
                               const std::string & year)
{
  AutoConnection conn;
  DBQuery query(conn);
  StrQuery str_query; 

  str_query.addSelect(concatas(DB::PLANT_TABLE_NAME, DB::ID));
  str_query.addSelect(concatas(DB::PLANT_TABLE_NAME, DB::USED_CAPACITY));

  str_query.addFrom(DB::PRODUCT_TABLE_NAME);
  str_query.addFrom(DB::PLANT_TABLE_NAME);

  str_query.addWhere(op(concat(DB::PRODUCT_TABLE_NAME, DB::ID),
                               id_to_str(product_id)));
  str_query.addWhere(op(concat(DB::PLANT_TABLE_NAME, DB::ID),
                        concat(DB::PRODUCT_TABLE_NAME, DB::PLANT_ID)));

  if (not query.exec(str_query))
    {
      std::string msg = "Cannot execute query: " + std::string(str_query);
      throw std::logic_error(msg);
    }

  if (not query.next())
    return 0;

  db_id_t plant_id = std::atol(query.getValue(0));
  real plant_capacity = std::atof(query.getValue(1));

  query.clear();
  str_query.clear();

  str_query.addSelect(concatas(DB::PRODUCT_TABLE_NAME, DB::ID));
  str_query.addSelect(concatas(DB::PRODUCTION_QUANTITIES_TABLE_NAME,
                               DB::PRODUCED_QUANTITY));
  str_query.addSelect(concatas(DB::PRODUCTION_QUANTITIES_TABLE_NAME,
                               DB::PRICE));

  str_query.addFrom(DB::PRODUCT_TABLE_NAME);
  str_query.addFrom(DB::PRODUCTION_QUANTITIES_TABLE_NAME);

  str_query.addWhere(op(concat(DB::PRODUCT_TABLE_NAME, DB::PLANT_ID),
                        id_to_str(plant_id)));
  str_query.addWhere(op(concat(DB::PRODUCT_TABLE_NAME, DB::ID),
                        concat(DB::PRODUCTION_QUANTITIES_TABLE_NAME,
                               DB::PRODUCT_ID)));
  str_query.addWhere(in_year(DB::YEAR, year));

  if (not query.exec(str_query))
    {
      std::string msg = "Cannot execute query: " + std::string(str_query);
      throw std::logic_error(msg);
    }

  real total_amount = 0.0;

  real amount = 0.0;

  while (query.next())
    {
      db_id_t p_id  = std::atol(query.getValue(0));
      real quantity = std::atof(query.getValue(1));
      real price    = std::atof(query.getValue(2));

      total_amount += quantity * price;

      if (p_id == product_id)
        amount = quantity * price;
    }

  real proportion = total_amount == 0.0 ? 0.0 : amount / total_amount;

  return plant_capacity * proportion;
}

std::tuple<real, real, real>
get_product_staff_numbers(const std::string & rif,
                          const std::string & year,
                          const db_id_t & product_id)
{
  AutoConnection conn;
  DBQuery query(conn);
  StrQuery str_query;

  /* Se le suma una unidad al año porque el registro de empleos se hace al
     año siguiente.
  */
  int y = std::atoi(year.c_str()) + 1;

  std::stringstream sstr_year;

  sstr_year << y;

  std::string max_date = sstr_year.str() + "-12-31";

  str_query.addSelect(DB::ADMINISTRATIVE);
  str_query.addSelect(DB::OPERATING);
  str_query.addSelect(DB::WORKDAY);

  str_query.addFrom(DB::COMPANY_OCCUPANCY_TABLE_NAME);

  str_query.addWhere(op(DB::COMPANY_RIF, single_quote(rif)));
  str_query.addWhere(op(DB::ACTIVE, "TRUE"));
  str_query.addWhere(op(DB::INSERTION_DATE, single_quote(max_date), "<="));

  str_query.addOrderBy(DB::INSERTION_DATE);

  str_query.setOrderByOption("DESC");

  if (not query.exec(str_query))
    {
      std::string msg = "Cannot execute query: " + std::string(str_query);
      throw std::logic_error(msg);
    }

  if (not query.next())
    {
      query.clear();
      str_query.clear();

      str_query.addSelect(DB::ADMINISTRATIVE);
      str_query.addSelect(DB::OPERATING);
      str_query.addSelect(DB::WORKDAY);

      str_query.addFrom(DB::COMPANY_OCCUPANCY_TABLE_NAME);

      str_query.addWhere(op(DB::COMPANY_RIF, single_quote(rif)));
      str_query.addWhere(op(DB::ACTIVE, "TRUE"));
      str_query.addWhere(op(DB::INSERTION_DATE, single_quote(max_date), ">"));

      str_query.addOrderBy(DB::INSERTION_DATE);

      str_query.setOrderByOption("ASC");

      if (not query.exec(str_query))
        {
          std::string msg = "Cannot execute query: " + std::string(str_query);
          throw std::logic_error(msg);
        }

      if (not query.next())
        return std::tuple<real, real, real>(0.0, 0.0, 0.0);
    }

  real administrative = std::atof(query.getValue(0));
  real operating = std::atof(query.getValue(1));
  real workday = std::atof(query.getValue(2));

  query.clear();
  str_query.clear();

  str_query.addSelect(concatas(DB::PRODUCT_TABLE_NAME, DB::ID));
  str_query.addSelect(concatas(DB::PRODUCTION_QUANTITIES_TABLE_NAME,
                               DB::PRODUCED_QUANTITY));
  str_query.addSelect(concatas(DB::PRODUCTION_QUANTITIES_TABLE_NAME,
                               DB::PRICE));

  str_query.addFrom(DB::PRODUCT_TABLE_NAME);
  str_query.addFrom(DB::PLANT_TABLE_NAME);
  str_query.addFrom(DB::PRODUCTION_QUANTITIES_TABLE_NAME);

  str_query.addWhere(op(concat(DB::PLANT_TABLE_NAME, DB::UE_RIF),
                        single_quote(rif)));
  str_query.addWhere(op(concat(DB::PLANT_TABLE_NAME, DB::ID),
                        concat(DB::PRODUCT_TABLE_NAME, DB::PLANT_ID)));
  str_query.addWhere(op(concat(DB::PRODUCT_TABLE_NAME, DB::ID),
                        concat(DB::PRODUCTION_QUANTITIES_TABLE_NAME,
                               DB::PRODUCT_ID)));
  str_query.addWhere(in_year(DB::YEAR, year));

  if (not query.exec(str_query))
    {
      std::string msg = "Cannot execute query: " + std::string(str_query);
      throw std::logic_error(msg);
    }

  real total_amount = 0.0;

  real amount = 0.0;

  while (query.next())
    {
      db_id_t p_id  = std::atol(query.getValue(0));
      real quantity = std::atof(query.getValue(1));
      real price    = std::atof(query.getValue(2));

      total_amount += quantity * price;

      if (p_id == product_id)
        amount = quantity * price;
    }

  real proportion = total_amount == 0.0 ? 0.0 : amount / total_amount;

  administrative *= proportion;
  operating *= proportion;

  return std::tuple<real, real, real>(workday, administrative, operating);
}

real get_external_sales(const std::string & tariff_code,
                        const std::string & seller_rif,
                        const std::string & year,
                        const std::string & buyers_to_exclude)
{
  AutoConnection conn;
  DBQuery query(conn);
  StrQuery str_query;

  str_query.addSelect(concatas(DB::INPUT_SUPPLIER_TABLE_NAME,
                               DB::BOUGHT_QUANTITY));
  str_query.addSelect(concatas(DB::ACQ_PRICE_TABLE_NAME, DB::NAC_ACQ_PRICE));

  str_query.addFrom(DB::INPUT_TABLE_NAME);
  str_query.addFrom(DB::INPUT_SUPPLIER_TABLE_NAME);
  str_query.addFrom(DB::SUPPLIER_TABLE_NAME);
  str_query.addFrom(DB::ACQ_PRICE_TABLE_NAME);
  str_query.addFrom(DB::PLANT_TABLE_NAME);

  str_query.addWhere(op(concat(DB::INPUT_TABLE_NAME, DB::TARIFF_CODE),
                        single_quote(tariff_code)));
  str_query.addWhere(op(concat(DB::INPUT_TABLE_NAME, DB::ID),
                        concat(DB::INPUT_SUPPLIER_TABLE_NAME,
                               DB::INPUT_ID)));
  str_query.addWhere(op(concat(DB::INPUT_SUPPLIER_TABLE_NAME,
                               DB::YEAR_R), year));
  str_query.addWhere(op(concat(DB::INPUT_SUPPLIER_TABLE_NAME,
                               DB::SUPPLIER_ID),
                        concat(DB::SUPPLIER_TABLE_NAME, DB::ID)));
  str_query.addWhere(op(concat(DB::SUPPLIER_TABLE_NAME, DB::RIF),
                        single_quote(seller_rif)));
  str_query.addWhere(op(concat(DB::ACQ_PRICE_TABLE_NAME, DB::INPUT_ID),
                        concat(DB::INPUT_TABLE_NAME, DB::ID)));
  str_query.addWhere(in_year(DB::YEAR, year));

  str_query.addWhere(op(concat(DB::INPUT_TABLE_NAME, DB::PLANT_ID),
                        concat(DB::PLANT_TABLE_NAME, DB::ID)));

  if (buyers_to_exclude.size() > 0)
    {

      std::string exclude = concat(DB::PLANT_TABLE_NAME, DB::UE_RIF) +
                            " NOT IN (" + buyers_to_exclude + ")";

      str_query.addWhere(exclude);
    }

  if (not query.exec(str_query))
    {
      std::string msg = "Cannot execute query: " + std::string(str_query);
      throw std::logic_error(msg);
    }

  real external_sales = 0.0;

  while (query.next())
    {
      external_sales += std::atof(query.getValue(0)) *
                        std::atof(query.getValue(1));
    }

  return external_sales;
}

